from enum import Enum

import pandas as pd

from psyke.tuning import Objective, SKEOptimizer
from psyke.tuning.orchid import OrCHiD
from psyke.utils import Target


class CRASH(SKEOptimizer):
    class Algorithm(Enum):
        ExACT = 1,
        CREAM = 2

    def __init__(self, predictor, dataframe: pd.DataFrame, max_error_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, max_depth: int = 10,
                 max_gauss_components: int = 5, patience: int = 5, output: Target = Target.CONSTANT,
                 objective: Objective = Objective.MODEL, normalization=None, discretization=None):
        super().__init__(predictor, dataframe, max_error_increase, min_rule_decrease, readability_tradeoff,
                         patience, objective, output, normalization, discretization)
        self.max_depth = max_depth
        self.max_gauss_components = max_gauss_components

    def search(self):
        self.params = []
        for algorithm in [OrCHiD.Algorithm.ExACT, OrCHiD.Algorithm.CREAM]:
            self.params += self.__search_algorithm(algorithm)

    def __search_algorithm(self, algorithm):
        params = []
        best = None

        for gauss_components in range(2, self.max_gauss_components + 1):
            data = self.dataframe.sample(n=gauss_components * 100) if gauss_components * 100 < len(self.dataframe) \
                else self.dataframe
            current_params = self.__search_components(data, algorithm, gauss_components)
            current_best = self._best(current_params)[1]
            if best is not None and self._score(best) <= self._score(current_best):
                break
            best = current_best
            params += current_params

        return params

    def __search_components(self, data, algorithm, gauss_components):
        orchid = OrCHiD(data, algorithm, self.output, self.max_error_increase, self.min_rule_decrease,
                        self.readability_tradeoff, self.patience, self.max_depth, gauss_components,
                        self.normalization, self.discretization)
        orchid.search()
        return [(*p, gauss_components, algorithm) for p in orchid.params]

    def _print_params(self, name, params):
        print("*****************************")
        print(f"Best {name}")
        print("*****************************")
        print(f"MAE = {params[0]:.2f}, {params[1]} rules")
        print(f"Algorithm = {params[5]}")
        print(f"Threshold = {params[3]:.2f}")
        print(f"Depth = {params[2]}")
        print(f"Gaussian components = {params[4]}")



import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

from psyke import get_default_random_seed, Target
from psyke.extraction.hypercubic import Grid, HyperCube, GenericCube, ClassificationCube, RegressionCube
from psyke.extraction.hypercubic.gridex import GridEx


class HEx(GridEx):
    """
    Explanator implementing HEx algorithm.
    """

    def __init__(self, predictor, grid: Grid, min_examples: int, threshold: float, output: Target = Target.CONSTANT,
                 discretization=None, normalization=None, seed: int = get_default_random_seed()):
        super().__init__(predictor, grid, min_examples, threshold, output, discretization, normalization, seed)
        self._default_surrounding_cube = True

    def _different_output(self, this_cube: GenericCube, other_cube: GenericCube) -> bool:
        if isinstance(this_cube.output, str) and this_cube.output == other_cube.output:
            return False
        if isinstance(this_cube.output, float) and abs(this_cube.output - other_cube.output) < self.threshold:
            return False
        if isinstance(this_cube.output, LinearRegression):
            raise NotImplementedError
        return True

    def _gain(self, parent_cube: GenericCube, new_cube: GenericCube) -> float:
        if isinstance(parent_cube, ClassificationCube):
            return parent_cube.output != new_cube.output
        return abs(parent_cube.error - new_cube.error) > self.threshold * .5

    def _iterate(self, surrounding: HyperCube, dataframe: pd.DataFrame):
        fake = dataframe.copy()
        surrounding.update(dataframe, self.predictor)
        prev = [surrounding]

        for iteration in self.grid.iterate():
            next_iteration = []
            for cube in prev:
                subcubes, fake = self._cubes_to_split(cube, surrounding, iteration, dataframe, fake, True)
                cleaned = [c for c in subcubes if c.count(dataframe) > 0 and self._gain(cube, c)]
                if len(subcubes) > len(cleaned):
                    if len(cleaned) > 0:
                        idx = np.any([c.filter_indices(fake.iloc[:, :-1]) for c in cleaned], axis=0)
                        cube.update(fake[~idx], self.predictor)
                    self._hypercubes = [cube] + self._hypercubes
                next_iteration += self._merge(cleaned, fake)
            prev = next_iteration.copy()
        self._hypercubes = [cube for cube in prev] + self._hypercubes



from __future__ import annotations

from abc import ABC
from enum import Enum

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, f1_score, accuracy_score, \
    adjusted_rand_score, adjusted_mutual_info_score, v_measure_score, fowlkes_mallows_score

from psyke.schema import DiscreteFeature
from psyke.utils import get_default_random_seed, Target, get_int_precision
from tuprolog.theory import Theory
from typing import Iterable
import logging

logging.basicConfig(level=logging.WARN)
logger = logging.getLogger('psyke')


class EvaluableModel(object):
    class Task(Enum):
        CLASSIFICATION = 1,
        REGRESSION = 2,
        CLUSTERING = 3

    class Score(Enum):
        pass

    class ClassificationScore(Score):
        ACCURACY = 1
        F1 = 2,
        INVERSE_ACCURACY = 3

    class RegressionScore(Score):
        MAE = 1
        MSE = 2
        R2 = 3

    class ClusteringScore(Score):
        ARI = 1,
        AMI = 2,
        V = 3,
        FMI = 4

    def __init__(self, discretization=None, normalization=None):
        self.discretization = discretization
        self.normalization = normalization

    def predict(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None) -> Iterable:
        """
        Predicts the output values of every sample in dataset.

        :param dataframe: is the set of instances to predict.
        :param mapping: for one-hot encoding.
        :return: a list of predictions.
        """
        return self.__convert(self._predict(dataframe), mapping)

    def _predict(self, dataframe: pd.DataFrame) -> Iterable:
        raise NotImplementedError('predict')

    def __convert(self, ys: Iterable, mapping: dict[str: int] = None) -> Iterable:
        if mapping is not None:
            inverse_mapping = {v: k for k, v in mapping.items()}
            ys = [inverse_mapping[y] for y in ys]
        if self.normalization is not None:
            m, s = self.normalization[list(self.normalization.keys())[-1]]
            ys = [prediction if prediction is None else prediction * s + m for prediction in ys]
        return ys

    def brute_predict(self, dataframe: pd.DataFrame, criterion: str = 'corner', n: int = 2,
                      mapping: dict[str: int] = None) -> Iterable:
        return self.__convert(self._brute_predict(dataframe, criterion, n, mapping), mapping)

    def _brute_predict(self, dataframe: pd.DataFrame, criterion: str = 'corner', n: int = 2,
                       mapping: dict[str: int] = None) -> Iterable:
        raise NotImplementedError('brute_predict')

    def unscale(self, values, name):
        if self.normalization is None or isinstance(values, LinearRegression):
            return values
        if isinstance(values, Iterable):
            values = [None if value is None else
                      value * self.normalization[name][1] + self.normalization[name][0] for value in values]
        else:
            values = values * self.normalization[name][1] + self.normalization[name][0]
        return values

    def score(self, dataframe: pd.DataFrame, predictor=None, fidelity: bool = False, completeness: bool = True,
              brute: bool = False, criterion: str = 'corners', n: int = 2,
              task: EvaluableModel.Task = Task.CLASSIFICATION,
              scoring_function: Iterable[EvaluableModel.Score] = [ClassificationScore.ACCURACY]):
        extracted = np.array(
            self.predict(dataframe.iloc[:, :-1]) if not brute else
            self.brute_predict(dataframe.iloc[:, :-1], criterion, n)
        )
        idx = [prediction is not None for prediction in extracted]
        y_extracted = extracted[idx]
        true = [dataframe.iloc[idx, -1]]

        if fidelity:
            if predictor is None:
                raise ValueError("Predictor must be not None to measure fidelity")
            true.append(predictor.predict(dataframe.iloc[idx, :-1]).flatten())

        if task == EvaluableModel.Task.REGRESSION:
            y_extracted = self.unscale(y_extracted, dataframe.columns[-1])
            true = [self.unscale(t, dataframe.columns[-1]) for t in true]

        res = {
                  score: EvaluableModel.__evaluate(true, y_extracted, score) for score in scoring_function
              }, sum(idx) / len(idx)
        return res if completeness else res[0]

    @staticmethod
    def __evaluate(y, y_hat, scoring_function):
        if scoring_function == EvaluableModel.ClassificationScore.ACCURACY:
            f = accuracy_score
        elif scoring_function == EvaluableModel.ClassificationScore.F1:
            def f(true, pred):
                return f1_score(true, pred, average='weighted')
        elif scoring_function == EvaluableModel.ClassificationScore.INVERSE_ACCURACY:
            def f(true, pred):
                return 1 - accuracy_score(true, pred)
        elif scoring_function == EvaluableModel.RegressionScore.R2:
            f = r2_score
        elif scoring_function == EvaluableModel.RegressionScore.MAE:
            f = mean_absolute_error
        elif scoring_function == EvaluableModel.RegressionScore.MSE:
            f = mean_squared_error
        elif scoring_function == EvaluableModel.ClusteringScore.ARI:
            f = adjusted_rand_score
        elif scoring_function == EvaluableModel.ClusteringScore.AMI:
            f = adjusted_mutual_info_score
        elif scoring_function == EvaluableModel.ClusteringScore.V:
            f = v_measure_score
        elif scoring_function == EvaluableModel.ClusteringScore.FMI:
            f = fowlkes_mallows_score
        else:
            raise ValueError("Scoring function not supported")
        return [f(yy, y_hat) for yy in y]


class Extractor(EvaluableModel, ABC):
    """
    An explanator capable of extracting rules from trained black box.

    Parameters
    ----------
    predictor : the underling black box predictor.
    discretization : A collection of sets of discretised features.
        Each set corresponds to a set of features derived from a single non-discrete feature.
    """

    def __init__(self, predictor, discretization: Iterable[DiscreteFeature] = None, normalization=None):
        super().__init__(normalization)
        self.predictor = predictor
        self.discretization = [] if discretization is None else list(discretization)

    def extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        """
        Extracts rules from the underlying predictor.

        :param dataframe: is the set of instances to be used for the extraction.
        :param mapping: for one-hot encoding.
        :param sort: alphabetically sort the variables of the head of the rules.
        :return: the theory created from the extracted rules.
        """
        raise NotImplementedError('extract')

    def mae(self, dataframe: pd.DataFrame, predictor=None, brute: bool = False, criterion: str = 'center',
            n: int = 3) -> float:
        """
        Calculates the predictions' MAE w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the mean absolute error.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :param brute: if True, a brute prediction is executed.
        :param criterion: creterion for brute prediction.
        :param n: number of points for brute prediction with 'perimeter' criterion.
        :return: the mean absolute error (MAE) of the predictions.
        """
        return self.score(dataframe, predictor, predictor is not None, False, brute, criterion, n,
                          Extractor.Task.REGRESSION, [Extractor.RegressionScore.MAE])[Extractor.RegressionScore.MAE][-1]

    def mse(self, dataframe: pd.DataFrame, predictor=None, brute: bool = False, criterion: str = 'center',
            n: int = 3) -> float:
        """
        Calculates the predictions' MSE w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the mean squared error.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :param brute: if True, a brute prediction is executed.
        :param criterion: creterion for brute prediction.
        :param n: number of points for brute prediction with 'perimeter' criterion.
        :return: the mean squared error (MSE) of the predictions.
        """
        return self.score(dataframe, predictor, predictor is not None, False, brute, criterion, n,
                          Extractor.Task.REGRESSION, [Extractor.RegressionScore.MSE])[Extractor.RegressionScore.MSE][-1]

    def r2(self, dataframe: pd.DataFrame, predictor=None, brute: bool = False, criterion: str = 'center',
            n: int = 3) -> float:
        """
        Calculates the predictions' R2 score w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the R2 score.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :param brute: if True, a brute prediction is executed.
        :param criterion: creterion for brute prediction.
        :param n: number of points for brute prediction with 'perimeter' criterion.
        :return: the R2 score of the predictions.
        """
        return self.score(dataframe, predictor, predictor is not None, False, brute, criterion, n,
                          Extractor.Task.REGRESSION, [Extractor.RegressionScore.R2])[Extractor.RegressionScore.R2][-1]

    def accuracy(self, dataframe: pd.DataFrame, predictor=None, brute: bool = False, criterion: str = 'center',
            n: int = 3) -> float:
        """
        Calculates the predictions' accuracy classification score w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the accuracy classification score.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :param brute: if True, a brute prediction is executed.
        :param criterion: creterion for brute prediction.
        :param n: number of points for brute prediction with 'perimeter' criterion.
        :return: the accuracy classification score of the predictions.
        """
        return self.score(dataframe, predictor, predictor is not None, False, brute, criterion, n,
                          Extractor.Task.CLASSIFICATION,
                          [Extractor.ClassificationScore.ACCURACY])[Extractor.ClassificationScore.ACCURACY][-1]

    def f1(self, dataframe: pd.DataFrame, predictor=None, brute: bool = False, criterion: str = 'center',
            n: int = 3) -> float:
        """
        Calculates the predictions' F1 score w.r.t. the instances given as input.

        :param dataframe: is the set of instances to be used to calculate the F1 score.
        :param predictor: if provided, its predictions on the dataframe are taken instead of the dataframe instances.
        :param brute: if True, a brute prediction is executed.
        :param criterion: creterion for brute prediction.
        :param n: number of points for brute prediction with 'perimeter' criterion.
        :return: the F1 score of the predictions.
        """
        return self.score(dataframe, predictor, predictor is not None, False, brute, criterion, n,
                          Extractor.Task.CLASSIFICATION,
                          [Extractor.ClassificationScore.F1])[Extractor.ClassificationScore.F1][-1]

    @staticmethod
    def cart(predictor, max_depth: int = 3, max_leaves: int = 3,
             discretization: Iterable[DiscreteFeature] = None, normalization=None, simplify: bool = True) -> Extractor:
        """
        Creates a new Cart extractor.
        """
        from psyke.extraction.cart import Cart
        return Cart(predictor, max_depth, max_leaves, discretization=discretization, normalization=normalization,
                    simplify=simplify)

    @staticmethod
    def divine(predictor, k: int = 5, patience: int = 15, close_to_center: bool = True,
               discretization: Iterable[DiscreteFeature] = None, normalization=None,
               seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new DiViNE extractor.
        """
        from psyke.extraction.hypercubic.divine import DiViNE
        return DiViNE(predictor, k=k, patience=patience, close_to_center=close_to_center,
                      discretization=discretization, normalization=normalization, seed=seed)

    @staticmethod
    def cosmik(predictor, max_components: int = 4, k: int = 5, patience: int = 15, close_to_center: bool = True,
               output: Target = Target.CONSTANT, discretization: Iterable[DiscreteFeature] = None, normalization=None,
               seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new COSMiK extractor.
        """
        from psyke.extraction.hypercubic.cosmik import COSMiK
        return COSMiK(predictor, max_components=max_components, k=k, patience=patience, close_to_center=close_to_center,
                      output=output, discretization=discretization, normalization=normalization, seed=seed)

    @staticmethod
    def iter(predictor, min_update: float = 0.1, n_points: int = 1, max_iterations: int = 600, min_examples: int = 250,
             threshold: float = 0.1, fill_gaps: bool = True, normalization: dict[str, tuple[float, float]] = None,
             output=None, seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new ITER extractor.
        """
        from psyke.extraction.hypercubic.iter import ITER
        return ITER(predictor, min_update, n_points, max_iterations, min_examples, threshold, fill_gaps,
                    normalization, output, seed)

    @staticmethod
    def gridex(predictor, grid, min_examples: int = 250, threshold: float = 0.1, output: Target = Target.CONSTANT,
               discretization=None, normalization: dict[str, tuple[float, float]] = None,
               seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new GridEx extractor.
        """
        from psyke.extraction.hypercubic.gridex import GridEx
        return GridEx(predictor, grid, min_examples, threshold, output, discretization, normalization, seed)

    @staticmethod
    def hex(predictor, grid, min_examples: int = 250, threshold: float = 0.1, output: Target = Target.CONSTANT,
            discretization=None, normalization: dict[str, tuple[float, float]] = None,
            seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new HEx extractor.
        """
        from psyke.extraction.hypercubic.hex import HEx
        return HEx(predictor, grid, min_examples, threshold, output, discretization, normalization, seed)

    @staticmethod
    def gridrex(predictor, grid, min_examples: int = 250, threshold: float = 0.1,
                normalization: dict[str, tuple[float, float]] = None,
                seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new GridREx extractor.
        """
        from psyke.extraction.hypercubic.gridrex import GridREx
        return GridREx(predictor, grid, min_examples, threshold, normalization, seed)

    @staticmethod
    def creepy(predictor, clustering, depth: int, error_threshold: float, output: Target = Target.CONSTANT,
               gauss_components: int = 2, ranks: [(str, float)] = [], ignore_threshold: float = 0.0,
               discretization=None, normalization: dict[str, tuple[float, float]] = None,
               seed: int = get_default_random_seed()) -> Extractor:
        """
        Creates a new CReEPy extractor.
        """
        from psyke.extraction.hypercubic.creepy import CReEPy
        return CReEPy(predictor, clustering, depth, error_threshold, output, gauss_components, ranks, ignore_threshold,
                      discretization, normalization, seed)

    @staticmethod
    def real(predictor, discretization=None) -> Extractor:
        """
        Creates a new REAL extractor.
        """
        from psyke.extraction.real import REAL
        return REAL(predictor, [] if discretization is None else discretization)

    @staticmethod
    def trepan(predictor, discretization=None, min_examples: int = 0, max_depth: int = 3,
               split_logic=None) -> Extractor:
        """
        Creates a new Trepan extractor.
        """
        from psyke.extraction.trepan import Trepan, SplitLogic
        if split_logic is None:
            split_logic = SplitLogic.DEFAULT
        return Trepan(predictor, [] if discretization is None else discretization, min_examples, max_depth, split_logic)


class Clustering(EvaluableModel, ABC):
    def __init__(self, discretization=None, normalization=None):
        super().__init__(discretization, normalization)

    def fit(self, dataframe: pd.DataFrame):
        raise NotImplementedError('fit')

    def explain(self):
        raise NotImplementedError('explain')

    @staticmethod
    def exact(depth: int = 2, error_threshold: float = 0.1, output: Target = Target.CONSTANT, gauss_components: int = 2,
              discretization=None, normalization=None, seed: int = get_default_random_seed()) -> Clustering:
        """
        Creates a new ExACT instance.
        """
        from psyke.clustering.exact import ExACT
        return ExACT(depth, error_threshold, output, gauss_components, discretization, normalization, seed)

    @staticmethod
    def cream(depth: int = 2, error_threshold: float = 0.1, output: Target = Target.CONSTANT, gauss_components: int = 2,
              discretization=None, normalization=None, seed: int = get_default_random_seed()) -> Clustering:
        """
        Creates a new CREAM instance.
        """
        from psyke.clustering.cream import CREAM
        return CREAM(depth, error_threshold, output, gauss_components, discretization, normalization, seed)



from __future__ import annotations

from collections import Iterable
import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from tuprolog.theory import Theory
from psyke import Clustering
from psyke.clustering import HyperCubeClustering
from psyke.extraction.hypercubic import HyperCubeExtractor
from psyke.utils import Target, get_default_random_seed


class CReEPy(HyperCubeExtractor):
    """
    Explanator implementing CReEPy algorithm.
    """

    def __init__(self, predictor, clustering=Clustering.exact, depth: int = 3, error_threshold: float = 0.1,
                 output: Target = Target.CONSTANT, gauss_components: int = 5, ranks: list[(str, float)] = [],
                 ignore_threshold: float = 0.0, discretization=None, normalization=None,
                 seed: int = get_default_random_seed()):
        super().__init__(predictor, Target.CLASSIFICATION if isinstance(predictor, ClassifierMixin) else output,
                         discretization, normalization)
        self.clustering = clustering(depth, error_threshold, self._output, gauss_components, discretization,
                                     normalization, seed)
        self.ranks = ranks
        self.ignore_threshold = ignore_threshold
        self._default_surrounding_cube = True

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        if not isinstance(self.clustering, HyperCubeClustering):
            raise TypeError("clustering must be a HyperCubeClustering")

        self.clustering.fit(dataframe)
        self._hypercubes = self.clustering.get_hypercubes()
        for cube in self._hypercubes:
            for dimension, relevance in self.ranks:
                if relevance < self.ignore_threshold:
                    cube[dimension] = [-np.inf, np.inf]
        return self._create_theory(dataframe)



from enum import Enum

import numpy as np
import pandas as pd

from psyke import Clustering, EvaluableModel
from psyke.tuning import Optimizer, IterativeOptimizer
from psyke.utils import Target


class OrCHiD(IterativeOptimizer):
    class Algorithm(Enum):
        ExACT = 1,
        CREAM = 2

    def __init__(self, dataframe: pd.DataFrame, algorithm, output: Target = Target.CONSTANT,
                 max_error_increase: float = 1.2, min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1,
                 patience: int = 5, max_depth: int = 10, gauss_components=10, normalization=None, discretization=None):
        super().__init__(dataframe, max_error_increase, min_rule_decrease, readability_tradeoff, max_depth, patience,
                         output, normalization, discretization)
        self.algorithm = algorithm
        self.gauss_components = gauss_components

    def search(self):
        self.params = self.__search_depth()

    def __search_depth(self):
        params, best = [], None

        for depth in range(1, self.max_depth + 1):
            current_params = self.__search_threshold(depth)
            current_best = self._best(current_params)[1]
            print()
            best, to_break = self._check_iteration_improvement(best, current_best)
            params += current_params

            if len(params) > 1 and to_break:
                break
        return params

    def __search_threshold(self, depth):
        step = 1.0
        threshold = 1.0
        params = []
        patience = self.patience
        while patience > 0:
            print(f"{self.algorithm}. Depth: {depth}. Threshold = {threshold:.2f}. "
                  f"Gaussian components = {self.gauss_components}. ", end="")
            clustering = (Clustering.cream if self.algorithm == OrCHiD.Algorithm.CREAM else Clustering.exact)(
                depth=depth, error_threshold=threshold, gauss_components=self.gauss_components, output=self.output
            )
            clustering.fit(self.dataframe)
            task, metric = \
                (EvaluableModel.Task.CLASSIFICATION, EvaluableModel.ClassificationScore.INVERSE_ACCURACY) \
                if self.output == Target.CLASSIFICATION else \
                (EvaluableModel.Task.REGRESSION, EvaluableModel.RegressionScore.MAE)
            p, n = clustering.score(self.dataframe, None, False, False, task=task,
                                    scoring_function=[metric])[metric][0], clustering.n_rules

            print(f"Predictive loss = {p:.2f}, {n} rules")

            if len(params) == 0:
                params.append((p, n, depth, threshold))
                threshold = p / 20
                step = p / self.patience * 0.75
                continue

            if (n == 1) or (p == 0.0):
                params.append((p, n, depth, threshold))
                break

            if p > params[0][0] * self.max_error_increase:
                break

            improvement = (params[-1][0] / p) + (1 - n / params[-1][1])

            if improvement <= 1 or n > np.ceil(params[-1][1] * self.min_rule_decrease):
                patience -= 1
            if p != params[-1][0] or n != params[-1][1]:
                params.append((p, n, depth, threshold))
            threshold += step
        return params

    def _print_params(self, name, params):
        print("*" * 40)
        print(f"* Best {name}")
        print("*" * 40)
        print(f"* Predictive loss = {params[0]:.2f}, {params[1]} rules")
        print(f"* Threshold = {params[3]:.2f}")
        print(f"* Depth = {params[2]}")
        print("*" * 40)



from __future__ import annotations

from typing import Iterable

import numpy as np
import pandas as pd

from psyke.utils import Target, get_default_random_seed
from psyke.clustering.exact import ExACT
from psyke.extraction.hypercubic import Node, HyperCube, ClosedCube
from psyke.clustering.utils import select_gaussian_mixture


class CREAM(ExACT):
    """
    Explanator implementing CREAM algorithm.
    """

    def __init__(self, depth: int, error_threshold: float, output: Target = Target.CONSTANT, gauss_components: int = 5,
                 discretization=None, normalization=None, seed: int = get_default_random_seed()):
        super().__init__(depth, error_threshold, output, gauss_components, discretization, normalization, seed)

    def __eligible_cubes(self, gauss_pred: np.ndarray, node: Node, clusters: int):
        cubes = []
        for i in range(len(np.unique(gauss_pred))):
            df = node.dataframe.iloc[np.where(gauss_pred == i)]
            if len(df) == 0:
                continue
            inner_cube = self._create_cube(df, clusters)
            indices = self._indices(inner_cube, node.dataframe)
            if indices is None:
                continue
            right, left = self._split(inner_cube, node.cube, node.dataframe, indices)
            cubes.append((
                ((right.diversity + left.diversity) / 2, right.volume(), left.volume(), i),
                (right, indices), (left, ~indices)
            ))
        return cubes

    def _split(self, right: ClosedCube, outer_cube: ClosedCube, data: pd.DataFrame, indices: np.ndarray):
        right.update(data.iloc[indices], self._predictor)
        left = outer_cube.copy()
        left.update(data.iloc[~indices], self._predictor)
        return right, left

    def _iterate(self, surrounding: Node) -> Iterable[HyperCube]:
        to_split = [(self.error_threshold * 10, 1, 1, surrounding)]
        while len(to_split) > 0:
            to_split.sort(reverse=True)
            (_, depth, _, node) = to_split.pop()
            data = ExACT._remove_string_label(node.dataframe)
            gauss_params = select_gaussian_mixture(data, self.gauss_components)
            gauss_pred = gauss_params[2].predict(data)
            cubes = self.__eligible_cubes(gauss_pred, node, gauss_params[1])
            if len(cubes) < 1:
                continue
            _, right, left = min(cubes)
            # find_better_constraints(node.dataframe[right[1]], right[0])
            node.right = Node(node.dataframe[right[1]], right[0])
            node.cube.update(node.dataframe[left[1]], self._predictor)
            node.left = Node(node.dataframe[left[1]], left[0])

            if depth < self.depth:
                to_split += [
                    (error, depth + 1, np.random.uniform(), n) for (n, error) in
                    zip(node.children, [right[0].diversity, left[0].diversity]) if error > self.error_threshold
                ]
        return self._node_to_cubes(surrounding)


from __future__ import annotations
import random as rnd
from itertools import product
from typing import Iterable
import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from tuprolog.theory import Theory
from psyke import get_default_random_seed
from psyke.utils import Target
from psyke.extraction.hypercubic import HyperCubeExtractor, Grid, HyperCube


class GridEx(HyperCubeExtractor):
    """
    Explanator implementing GridEx algorithm, doi:10.1007/978-3-030-82017-6_2.
    """

    def __init__(self, predictor, grid: Grid, min_examples: int, threshold: float, output: Target = Target.CONSTANT,
                 discretization=None, normalization=None, seed: int = get_default_random_seed()):
        super().__init__(predictor, Target.CLASSIFICATION if isinstance(predictor, ClassifierMixin) else output,
                         discretization, normalization)
        self.grid = grid
        self.min_examples = min_examples
        self.threshold = threshold
        self._generator = rnd.Random(seed)

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        self._hypercubes = []
        surrounding = HyperCube.create_surrounding_cube(dataframe, output=self._output)
        surrounding.init_diversity(2 * self.threshold)
        self._iterate(surrounding, dataframe)
        return self._create_theory(dataframe, sort)

    def _create_ranges(self, cube, iteration):
        ranges = {}
        for (feature, (a, b)) in cube.dimensions.items():
            n_bins = self.grid.get(feature, iteration)
            if n_bins == 1:
                ranges[feature] = [(-np.inf, np.inf)]
            else:
                size = (b - a) / n_bins
                ranges[feature] = [(a + size * i, a + size * (i + 1)) for i in range(n_bins)]
        return ranges

    def _cubes_to_split(self, cube, surrounding, iteration, dataframe, fake, keep_empty=False):
        to_split = []
        for (pn, p) in enumerate(list(product(*self._create_ranges(cube, iteration).values()))):
            cube = self._default_cube()
            for i, f in enumerate(dataframe.columns[:-1]):
                cube.update_dimension(f, p[i])
            n = cube.count(dataframe)
            if n > 0 or keep_empty:
                fake = pd.concat([fake, cube.create_samples(self.min_examples - n, surrounding, self._generator)])
                cube.update(fake, self.predictor)
                to_split.append(cube)
        return to_split, fake

    def _iterate(self, surrounding: HyperCube, dataframe: pd.DataFrame):
        fake = dataframe.copy()
        prev = [surrounding]
        next_iteration = []

        for iteration in self.grid.iterate():
            next_iteration = []
            for cube in prev:
                if cube.count(dataframe) == 0:
                    continue
                if cube.diversity < self.threshold:
                    self._hypercubes += [cube]
                    continue
                to_split, fake = self._cubes_to_split(cube, surrounding, iteration, dataframe, fake)
                next_iteration += [c for c in self._merge(to_split, fake)]
            prev = next_iteration.copy()
        self._hypercubes += [cube for cube in next_iteration]

    @staticmethod
    def _find_couples(to_split: Iterable[HyperCube], not_in_cache: Iterable[HyperCube],
                      adjacent_cache: dict[tuple[HyperCube, HyperCube], str | None]) -> \
            Iterable[tuple[HyperCube, HyperCube, str]]:
        checked = []
        eligible = []
        for cube in to_split:
            checked.append(cube)
            for other_cube in [c for c in to_split if c not in checked]:
                if (cube in not_in_cache) or (other_cube in not_in_cache):
                    adjacent_cache[(cube, other_cube)] = cube.is_adjacent(other_cube)
                adjacent_feature = adjacent_cache[(cube, other_cube)]
                eligible.append((cube, other_cube, adjacent_feature))
        return [couple for couple in eligible if couple[2] is not None]

    def _evaluate_merge(self, not_in_cache: Iterable[HyperCube],
                        dataframe: pd.DataFrame, feature: str,
                        cube: HyperCube, other_cube: HyperCube,
                        merge_cache: dict[(HyperCube, HyperCube), HyperCube | None]) -> bool:
        if (cube in not_in_cache) or (other_cube in not_in_cache):
            merged_cube = cube.merge_along_dimension(other_cube, feature)
            merged_cube.update(dataframe, self.predictor)
            merge_cache[(cube, other_cube)] = merged_cube
        return cube.output == other_cube.output if self._output == Target.CLASSIFICATION else \
            merge_cache[(cube, other_cube)].diversity < self.threshold

    def _merge(self, to_split: Iterable[HyperCube], dataframe: pd.DataFrame) -> Iterable[HyperCube]:
        not_in_cache = [cube for cube in to_split]
        adjacent_cache = {}
        merge_cache = {}
        cont = True
        while cont:
            to_merge = [([cube, other_cube], merge_cache[(cube, other_cube)]) for cube, other_cube, feature in
                        GridEx._find_couples(to_split, not_in_cache, adjacent_cache) if
                        self._evaluate_merge(not_in_cache, dataframe, feature, cube, other_cube, merge_cache)]
            if len(to_merge) == 0:
                cont = False
            else:
                sorted(to_merge, key=lambda c: c[1].diversity)
                best = to_merge[0]
                to_split = [cube for cube in to_split if cube not in best[0]] + [best[1]]
                not_in_cache = [best[1]]
        return to_split



from abc import ABC
from enum import Enum
import numpy as np
import pandas as pd

from psyke.extraction.hypercubic import Grid
from psyke.utils import Target


class Objective(Enum):
    MODEL = 1,
    DATA = 2


class Optimizer:
    def __init__(self, dataframe: pd.DataFrame, output: Target = Target.CONSTANT, max_error_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, patience: int = 5,
                 normalization=None, discretization=None):
        self.dataframe = dataframe
        self.output = output
        self.max_error_increase = max_error_increase
        self.min_rule_decrease = min_rule_decrease
        self.readability_tradeoff = readability_tradeoff
        self.patience = patience
        self.params = None
        self.normalization = normalization
        self.discretization = discretization

    def search(self):
        raise NotImplementedError

    def _best(self, params):
        param_dict = {self._score(t): t for t in params}
        min_param = min(param_dict)
        return min_param, param_dict[min_param]

    def _score(self, param):
        return param[0] * np.ceil(param[1] * self.readability_tradeoff)

    def _best_param(self, param):
        param_dict = {t[param]: t for t in self.params}
        min_param = min(param_dict)
        return min_param, param_dict[min_param]

    def get_best(self):
        names = ["Combined", "Predictive loss", "N rules"]
        params = [self._best(self.params), self._best_param(0), self._best_param(1)]
        for n, p in zip(names, params):
            self._print_params(n, p[1])
            print()
        return self._best(self.params)[1], self._best_param(0)[1], self._best_param(1)[1]

    def _print_params(self, n, param):
        raise NotImplementedError


class SKEOptimizer(Optimizer, ABC):
    def __init__(self, predictor, dataframe: pd.DataFrame, max_error_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, patience: int = 5,
                 objective: Objective = Objective.MODEL, output: Target = Target.CONSTANT,
                 normalization=None, discretization=None):
        super().__init__(dataframe, output, max_error_increase, min_rule_decrease, readability_tradeoff,
                         patience, normalization, discretization)
        self.predictor = predictor
        self.objective = objective


class IterativeOptimizer(Optimizer, ABC):
    def __init__(self, dataframe: pd.DataFrame, max_error_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, max_depth: int = 10,
                 patience: int = 5, output: Target = Target.CONSTANT, normalization=None, discretization=None):
        super().__init__(dataframe, output, max_error_increase, min_rule_decrease, readability_tradeoff,
                         patience, normalization, discretization)
        self.max_depth = max_depth

    def _iteration_improvement(self, best, other):
        if other[0] == best[0]:
            return (best[1] - other[1]) * 2
        return 1 / (
                (1 - other[0] / best[0]) ** self.readability_tradeoff *
                np.ceil(other[1] / self.readability_tradeoff) / np.ceil(best[1] / self.readability_tradeoff)
        )

    def _check_iteration_improvement(self, best, current):
        improvement = \
            self._iteration_improvement([best[0], best[1]], [current[0], current[1]]) if best is not None else np.inf
        return current, improvement < 1.2



from abc import ABC

import pandas as pd
from numpy import argmax
from tuprolog.theory import Theory

from psyke import Extractor


class PedagogicalExtractor(Extractor, ABC):

    def __init__(self, predictor, discretization=None, normalization=None):
        Extractor.__init__(self, predictor=predictor, discretization=discretization, normalization=normalization)

    def extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        from psyke.extraction.hypercubic import HyperCubeExtractor, HyperCube
        new_y = self.predictor.predict(dataframe.iloc[:, :-1])
        if mapping is not None:
            if hasattr(new_y[0], 'shape'):
                # One-hot encoding for multi-class tasks
                if len(new_y[0].shape) > 0 and new_y[0].shape[0] > 1:
                    new_y = [argmax(y, axis=0) for y in new_y]
                # One-hot encoding for binary class tasks
                else:
                    new_y = [round(y[0]) for y in new_y]
        new_y = pd.DataFrame(new_y).set_index(dataframe.index)
        data = dataframe.iloc[:, :-1].copy().join(new_y)
        data.columns = dataframe.columns
        return self._extract(data, mapping, sort)

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        raise NotImplementedError('extract')
